import { filter, interval, map, Observable, Observer, Subject, Subscription } from "rxjs";
import { v4 as uuidv4 } from 'uuid'
import config from '../config/config.json';
import { MessageTransmissionManager } from "../transmission/msg.transmission.manager";
import { WrappedMessage } from "../utils/message.ordering";
import ConsoleLogger from "../utils/log.utils";
import { FisMessage, GeneralEvent, TransmissionInterface, Transport, TransportMessage, TransportServiceInterface, TransportSet } from "../interface/interface";
import { WebsocketTransportService } from "../transport/websocket";
import { HttpTransportService } from "../transport/http";

class Supervisor {
    private generalBus: Subject<GeneralEvent<any>> = new Subject()
    private console = new ConsoleLogger('Supervisor', ['base'])
    private isClient: boolean = true
    private transmissionManager!: MessageTransmissionManager
    private event: Subject<GeneralEvent<any>>
    private transmissionSets: TransmissionInterface[] = []
    private outgoingPipe: Subject<any> = new Subject()
    private transportServiceArray: TransportServiceInterface[] = []
    private transportSet: TransportSet[] = []

    constructor() {
        this.event = new Subject()
        // Start setting up existing transport based on .env file
        this.sortTransportFromEnv(this.transportSet)
        this.transportSet.forEach(transport => {
            this.setUpTransportService(transport, this.event, this.isClient)
        })
        // once adapter manager is instantiated, it will attempt to connect to existing started transport
        this.tieInAdapterWithExistingTransportServices(this.event)

        this.transmissionManager = new MessageTransmissionManager(this.event, this.isClient)
        this.startMessageTransmission()
    }

    private startMessageTransmission(): void {
        // every new remote client connected, a new transmission object will be instantiated to allow message transmission
        this.transmissionManager.subscribeForTransmission().pipe(
            filter(event => event.type === `Transmission Event`),
            filter(event => event.event === `New Transmission`)
        ).subscribe((event: GeneralEvent<TransmissionInterface>) => {
            // broadcast to indicate a transmission object is ready, this signal will be received by adapter manager to instantiate the releavnt adapters to be used
            this.event.next({
                id: uuidv4(),
                type: `General Event`,
                event: `New Transmission`,
                date: new Date(),
                data: event.data
            })
            let transmission: TransmissionInterface = event.data
            this.console.log({ message: `Acquired transmission set for client ${transmission.clientId}` })
            // updating transmission records. No logic for keeping track of client state, as that will be handled exclusively by tranmistter transmission at the moment
            this.transmissionSets.push(transmission)

            this.handleClientActivity(transmission)
        })
    }

    private handleClientActivity(messageTransmission: TransmissionInterface): void {
        // start listening to incoming messages from this client
        messageTransmission.receiver.getIncoming().subscribe((event: GeneralEvent<any>) => {
            this.console.log({ message: `General Bus ${event.event} ${(((event.data as TransportMessage)?.payload as WrappedMessage)?.payload as FisMessage)?.header?.messageID ?? 'Not Message'}`, details: event })
            this.generalBus.next(event)
        })

        // for all the responses or messages to be emitted on the provider perspective
        this.outgoingPipe.subscribe((message: FisMessage) => {
            messageTransmission.transmitter.emit(message)
        })

        // test sample to simulate request response, but not using request response transmission
        let request: FisMessage = {
            header: {
                messageID: uuidv4(),
                messageName: 'RequestMessage'
            },
            data: 'Data'
        }
        // this.request(request, messageTransmission).subscribe({
        //     next: res => this.console.log({ message: `received ${res.header.messageID}`, details: res }),
        //     complete: () => this.console.log({ message: `Responses Completed for request: ${request.header.messageID}` })
        // })

        // test sample to stream new messages every second
        // this.startGeneratingRequest(1000, this.outgoingPipe)
    }

    private request(request: FisMessage, messageTransmission: TransmissionInterface): Observable<any> {
        return new Observable((response: Observer<any>) => {
            messageTransmission.transmitter.emit(request)
            this.generalBus.pipe(
                filter(event => event.event == 'New Message'),
                filter(event => (((event.data as TransportMessage)?.payload as WrappedMessage)?.payload as FisMessage)?.header.messageID === request.header.messageID),
                map(event => (((event.data as TransportMessage)?.payload as WrappedMessage)?.payload as FisMessage))
            ).subscribe(message => {
                if (message.data == 'Complete') {
                    response.complete()
                } else {
                    response.next(message)
                }
            })
        })
    }


    private startGeneratingRequest(intervalDuration: number, requestsPipe: Subject<FisMessage>) {
        interval(intervalDuration).subscribe(time => {
            let message: FisMessage = {
                header: {
                    messageID: uuidv4(),
                    messageName: 'NotificationMessage'
                },
                data: 'Data'
            }
            requestsPipe.next(message)
        })
    }

    private setUpTransportService(transportSet: TransportSet, event: Subject<GeneralEvent<any>>, isClient?: boolean): void {
        try {
            let transportService: TransportServiceInterface = this.instantiateTransportService(transportSet.transport, event)
            this.transportServiceArray.push(transportService)
            if (transportService instanceof WebsocketTransportService) {
                this.console.log({ message: `Just Double Checking... this is websocket` })
                if (isClient) {
                    // please note this is subject to change depending on the UI environemnt. Angular has their own built in function to read json file based on Swopt-UI
                    transportService.startClient(config.connection.transmitter)
                } else {
                    transportService.startServer(transportSet.port);
                }
            } else if (transportService instanceof HttpTransportService) {
                this.console.log({ message: `Just Double Checking... this is http` })
                // Additional Http-specific setup if needed.
                if (isClient) {
                    transportService.startClient(config.connection.transmitter)
                } else {
                    transportService.startServer(transportSet.port)
                }
            }
        } catch (error) {
            this.console.error({ message: 'Fail to set transport. Error in setting up transport', details: error })
        }
    }

    private instantiateTransportService(transportType: Transport, event: Subject<GeneralEvent<any>>): TransportServiceInterface {
        if (transportType === 'Websocket') {
            return new WebsocketTransportService(event)
        }
        else if (transportType === 'Http') {
            return new HttpTransportService(event)
        } else {
            throw new Error(`No Transport Service Instantiated`)
        }
    }

    // just to re-arrange the list of transport servicce from env. Of course, varying modules will have varying ways to start up their transport respectively. TO adapter later
    private sortTransportFromEnv(transportSet: TransportSet[]): void {
        let transportList: string[] = process.env.Transport?.split(',') || []
        let portList: number[] = (process.env.PORT?.split(',') || []).map(port => Number(port));
        transportList.forEach((transport, index) => {
            transportSet.push({ transport: transport, port: portList[index] } as unknown as TransportSet)
        })
        this.console.log({ message: 'TransportSetList', details: this.transportSet })
    }

    // A method to broadcast avaible tranpsort to adapterManager the available transport to be connected after the adapter manager has been instantiated
    private tieInAdapterWithExistingTransportServices(eventBus: Subject<GeneralEvent<any>>): void {
        const subscription: Subscription = eventBus.pipe(
            filter(event => event.type === `Adapter Event`),
            filter(event => event.event === `Adapter Manager Started`)
        ).subscribe(event => {
            eventBus.next({
                id: uuidv4(),
                type: `General Event`,
                event: `Available Transport`,
                date: new Date(),
                data: this.transportServiceArray
            })
        })
    }

}

let supervisor = new Supervisor()
